// RunCPM extensions added by P.Z.
#ifndef MYEXT_H
#define MYEXT_H
#define TPAADR (TPASIZE * 1024) // 0xf000 for RunCMP 60K
#define WORKADR 49152 // 0xc000

#ifdef CCP_ZCPR2
void _init_zcpr2()
{
/*
 * Memory map (62 k, TPAADR = 0xf800):
 * EC00-F3FF    CCP (ZCPR2)
 * F400-F7FF    BDOS
 * F800-F8CF    MCL buffer (200 + 4 + dead space bytes)
 * F8DO-F8FF    external stack (48 bytes)
 * F900-FB81    ndirs buffer (642 bytes, 64 dir entries)
 * FB82-FDFF    free space (638 bytes)
 * FE00-FFFF    BIOS
 */
	_puts("Initializing ZCPR2 buffers.\r\n");
	// Init of MCL buffer
	char startup[] = "STARTUP\0";
	_RamWrite16(TPAADR, TPAADR + 4); // address of first char in MCL buffer
	_RamWrite(TPAADR + 2, 200); // 200 bytes buffer length
	_RamStr(startup, TPAADR + 4); // init command into buffer
	// Init of external search path on address 0x0040
	_RamWrite(0x0040, 0); // no entry, path is set via STARTUP -> PATH
	// Init of ndirs buffer
	_RamWrite(TPAADR + 0x100, 64); // max. 64 entries
	_RamFill(TPAADR + 0x101, 4, 0); // no entry, ndirs are set via STARTUP -> LD
}
#endif

#ifdef CCP_ZCPR3
void _init_zcpr3()
{
	_puts("Initializing ZCPR33 buffers.\r\n");

	// Set all memory for buffers to zero
	_RamFill(TPAADR, BIOSjmppage - TPAADR, 0);

	// Init of Environment Descriptor
  _RamLoad((char *)"A/0/SYS.ENV", WORKADR);
  _RamCopy(WORKADR, 256, _RamRead16(WORKADR + 27)); // 27 = offset for Z3ENV

	// Init of Resident Command Package
  uint16 dest = _RamRead16(WORKADR + 12); // 12 = offset for RCP
  if (dest) {
  	_RamLoad((char *)"A/0/SYS.RCP", dest);
  }

	// Init of Flow Command Package
  dest = _RamRead16(WORKADR + 18); // 18 = offset for FCP
  if (dest) {
  	_RamLoad((char *)"A/0/SYS.FCP", dest);
  }

	// Init of Named Directory Buffer
  dest = _RamRead16(WORKADR + 21); // 21 = offset for Z3NDIRS
  if (dest) {
	  _RamLoad((char *)"A/0/SYS.NDR", dest);
	}

	// Init of Command Line Buffer
  _RamLoad((char *)"A/0/SYS.MCL", _RamRead16(WORKADR + 24)); // 24 = offset for Z3CL

  // Init of Wheel Byte
  _RamWrite(_RamRead16(WORKADR + 41), 255); // 41 = ofset for Z3WHL
}
#endif

#endif

